from random import choice as randchoice

def cache(func, **kwargs):
    """ Simple function caching decorator. """
    call = None
    try:
        from functools import lru_cache # Python 3 LRU cache = good
    except ImportError as e:
        # No Python 3
        maxsize = kwargs.get('maxsize', 32)
        cdict = {}
        def call(*args, **kwargs):
            keystruct = tuple(args), tuple(kwargs.items())
            key = hash(keystruct)
            if key in cdict:
                return cdict[key]
            newresult = func(*args, **kwargs)
            if len(cdict)>maxsize:
                del cdict[randchoice(cdict.keys())]
            cdict[key] = newresult
            return newresult
    else:
        call = lru_cache(**kwargs)(func)
    finally:
        return call
    



"""
Generic linux daemon base class for python 3.x.

Modified, originally from http://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python/
"""

import logging, sys, os, time, atexit, signal, traceback

class Daemon(object):
    """A generic daemon class.

    Usage: subclass the daemon class and override the run() method."""

    def __init__(self, pidfile):
        self.pidfile = pidfile
    
    def daemonize(self):
        """Deamonize class. UNIX double fork mechanism."""
        try: 
            pid = os.fork() 
            if pid > 0:
                # exit first parent
                sys.exit(0) 
        except OSError as err: 
            sys.stderr.write('fork #1 failed: {0}\n'.format(err))
            sys.exit(1)
    
        # decouple from parent environment
        os.chdir('/') 
        os.setsid() 
        os.umask(0) 
    
        # do second fork
        try: 
            pid = os.fork() 
            if pid > 0:

                # exit from second parent
                sys.exit(0) 
        except OSError as err: 
            sys.stderr.write('fork #2 failed: {0}\n'.format(err))
            sys.exit(1) 
    
        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        si = open(os.devnull, 'r')
        so = open(os.devnull, 'a+')
        se = open(os.devnull, 'a+')

        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())
    
        # write pidfile
        atexit.register(self.delpid)

        pid = str(os.getpid())
        with open(self.pidfile,'w+') as f:
            f.write(pid + '\n')
    
    def delpid(self):
        os.remove(self.pidfile)

    def start(self):
        """Start the daemon."""

        # Check for a pidfile to see if the daemon already runs
        try:
            with open(self.pidfile,'r') as pf:

                pid = int(pf.read().strip())
        except IOError:
            pid = None
    
        if pid:
            message = "pidfile {0} already exist. " + \
                    "Daemon already running?\n"
            sys.stderr.write(message.format(self.pidfile))
            sys.exit(1)
        
        # Start the daemon
        sys.stderr.write("Daemonizing!\n")
        self.daemonize()
        self.run()

    def stop(self):
        """Stop the daemon."""

        # Get the pid from the pidfile
        try:
            with open(self.pidfile,'r') as pf:
                pid = int(pf.read().strip())
        except IOError:
            pid = None
    
        if not pid:
            message = "pidfile {0} does not exist. " + \
                    "Daemon not running?\n"
            sys.stderr.write(message.format(self.pidfile))
            return # not an error in a restart

        # Try killing the daemon process    
        try:
            while 1:
                os.kill(pid, signal.SIGTERM)
                time.sleep(0.1)
        except OSError as err:
            e = str(err.args)
            if e.find("No such process") > 0:
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
            else:
                print (str(err.args))
                sys.exit(1)

    def restart(self):
        """Restart the daemon."""
        self.stop()
        self.start()

    def run(self):
        """You should override this method when you subclass Daemon.
        
        It will be called after the process has been daemonized by 
        start() or restart()."""
        raise NotImplementedError('Daemon.run() not implemented. Please subclass Daemon to implement it.')

def run_as_service(daemon):
    meths = {
        'start': daemon.start,
        'stop': daemon.stop,
        'restart': daemon.restart,
    }

    if len(sys.argv)<2 or sys.argv[1] not in ('start', 'stop', 'restart'):
        print("Please specify 'start', 'stop', or 'restart'")
    else:
        try:
            meths[sys.argv[1]]()
        except Exception as e:
            logging.error(sys.exc_info())
            logging.error(traceback.format_exc())


def iter_chunks(collection, chunksize):
    """
    iter_chunks(collection, chunksize) -> iterator
    
    Get an iterator that returns lists containing chunksize items,
    sequentially extracted from the given collection.
    """
    iterator = iter(collection)
    
    class OutOfInputException(Exception):
        pass
    
    def get_chunk():
        buffer = []
        counter = 0
        for value in iterator:
            buffer.append(value)
            counter += 1
            if counter >= chunksize:
                break
        if not counter:
            raise OutOfInputException()
        return buffer
    
    while True:
        try:
            yield get_chunk()
        except OutOfInputException:
            break

    


## Revamped daemon implementation based on python-daemon
#
# Usage #1 - turning a Python function into a service:
#
# serv = InitService('/path/to/lockfile', myfunc, fork=True)
# serv.run_cmdline()
#
# Usage #2 - turning another script into a service:
#
# serv = InitService('/path/to/lockfile', '/path/to/script', prog_name='script_alias')
# serv.run_cmdline()
#
# =======
#
# $ python my_script.py -h
# usage: my_script.py [-h] {start,stop,restart,reload,status}
#
# positional arguments:
#   {start,stop,restart,reload,status}
#                         service action to take
# 
# optional arguments:
#   -h, --help            show this help message and exit


import argparse, contextlib, daemon, os, psutil, subprocess, sys, time

# Include ExitStack context into contextlib for Python 2
try:
    from contextlib2 import ExitStack
    contextlib.ExitStack = ExitStack
except ImportError:
    pass

class InitService(object):
    def __init__(self, lockfile,
                 executable, args=[], prog_name=None,
                 verbose=False, fork=False,
                 cwd=None, env=None):
        self.lockfile = lockfile
        self.fork = fork
        self.cwd = cwd
        self.env = env

        self.prog_name = prog_name

        self.executable_str = None
        self.executable_py = None
        if hasattr(executable, '__call__'):
            # If callable, assume you were given a Python function
            self.executable_py = executable
            if self.prog_name:
                raise ValueError("Cannot specify prog_name when running a Python function")
        else:
            self.executable_str = str(executable)

        self.exec_args = args

        self.verbose_output = open(os.devnull, 'w')
        if self.verbose_output:
            if hasattr(verbose, 'write') and hasattr(verbose.write, '__call__'):
                self.verbose_output = verbose
            else:
                self.verbose_output = sys.stdout

    def _pid_from_name(self):
        if self.executable_py:
            return None # Name can't be used here
        name = self.prog_name or self.executable_str
        for proc in psutil.process_iter():
            if proc.cmdline() and proc.cmdline()[0] == name:
                return proc.pid
        return None
                
    @property
    def pid(self):
        returnpid = None
        
        if not os.path.exists(self.lockfile):
            return None
        try:
            with open(self.lockfile) as lock:
                pid = lock.read().strip()
            returnpid = int(pid)
        except Exception as e:
            # Error reading pid
            self.pid = None
            namepid = self._pid_from_name()
            if namepid:
                self.pid = namepid
            return namepid

        if returnpid not in psutil.pids():
            # No such process
            self.pid = None
            namepid = self._pid_from_name()
            self.pid = namepid
            return namepid

        return returnpid

    @pid.setter
    def pid(self, val):
        if not val:
            try:
                os.unlink(self.lockfile)
            except FileNotFoundError as e:
                pass
            finally:
                return
        with open(self.lockfile, 'w') as lock:
            lock.write("%s" % val)
                

    def run_cmdline(self, argv=[],
                    description="pyinit service manager"):
        parser = argparse.ArgumentParser()
        parser.add_argument('action', type=str, help="service action to take",
                            choices=['start', 'stop', 'restart', 'reload', 'status'])
        args = parser.parse_args(argv[1:] or sys.argv[1:])
        
        self.run_action(args.action)

    def run_action(self, action):
        action_map = {
            'start': self._start,
            'stop': self._stop,
            'restart': self._restart,
            'reload': self._reload,
            'status': self._status,
            }
        action_map[action]()
        
        
    def _start(self):
        self.pre_start()
        self.verbose_output.write('Starting service...\n')
        if self.pid:
            print('PID %s exists; process already running (clear %s if this is wrong)\n' % (self.pid, self.lockfile))
            return
        
        context = contextlib.ExitStack()
        if self.fork:
            context = daemon.DaemonContext(files_preserve=[self.verbose_output])
        
        if self.executable_str:
            # Use Popen to run the string command
            self.verbose_output.write("String command: %s\n" % self.executable_str)
            self.verbose_output.write("Run to look like command: %s\n" % self.prog_name)
            cmd = [self.executable_str] + self.exec_args
            executable = None
            if self.prog_name:
                cmd[0] = self.prog_name
                executable = self.executable_str

            proc = subprocess.Popen(cmd, executable=executable)
            self.pid = proc.pid
            self.verbose_output.write('Service started. %s\n' % self.pid)
        else:
            # Use daemon context to run the Python command
            self.verbose_output.write('Running %s(%s).\n' % (self.executable_py, self.exec_args))
            with context:
                print(os.getpid())
                self.pid = os.getpid()
                self.executable_py(*self.exec_args)
        self.post_start()
            
        
    def _stop(self):
        self.pre_stop()
        self.verbose_output.write('Stopping service...\n')
        pid = self.pid
        if not pid:
            self.verbose_output.write('Unknown PID; is the service running?\n')
            return
        proc = psutil.Process(self.pid)
        proc.terminate()
        self.verbose_output.write('Term signal sent.\n')
        self.post_stop()
        
    def _restart(self):
        self.verbose_output.write('Restarting service...\n')
        self._stop()
        time.sleep(1)
        self._start()

    def _reload(self):
        self.verbose_output.write('Reloading service...\n')
        self.reload()

    def _status(self):
        self.verbose_output.write('Status-ing service...\n')
        stat = None
        if self.pid:
            proc = psutil.Process(self.pid)
            stat = proc.status()
        else:
            stat = 'stopped'
        self.verbose_output.write('%s (pid: %s)\n' % (stat, self.pid))
        self.status(stat, self.pid)
        

    # User hooks
    def pre_start(self):
        pass
    def post_start(self):
        pass
    def pre_stop(self):
        pass
    def post_stop(self):
        pass
    def status(self, stmsg, pid):
        pass
    def reload(self):
        pass



import os

SERIAL = 'yaul.parallel.SERIAL'
THREADED = 'yaul.parallel.THREADED'
MULTIPROCESSED = 'yaul.parallel.MULTIPROCESSED'
PARALLEL_TYPES = [SERIAL, THREADED, MULTIPROCESSED]
_DEFAULT = SERIAL

def set_default_parallel(parallel_type):
    if not parallel_type in PARALLEL_TYPES:
        raise ValueError("Unrecognized parallelism type", parallel_type)
    global _DEFAULT
    _DEFAULT = parallel_type

class GenericPool(object):
    """Generic pool class supporting multiple kinds of parallelism."""
    
    def __init__(self, parallel_type=None, num_threads=None):
        if not parallel_type:
            parallel_type = _DEFAULT
        if not parallel_type in PARALLEL_TYPES:
            raise ValueError("Unrecognized parallelism type", parallel_type)
        self.partype = parallel_type
        self.num_threads = None
        self.pool = None

    def get_mapper(self):
        if self.partype == SERIAL:
            return self.__map_serial
        if self.partype == THREADED:
            return self.__map_threaded
        if self.partype == MULTIPROCESSED:
            return self.__map_multiprocessed
        raise ValueError("Unrecognized parallelism type", parallel_type)

    def map_reduce(self, mapfunc, collection, reducer=None, blocking=False):
        if not reducer:
            reducer = self.__null_reducer
        mapper = self.get_mapper()
        results = mapper(mapfunc, collection)
        reduced_results = reducer(results)
        if blocking:
            final_results = [x for x in reduced_results]
            return final_results
        else:
            return iter(reduced_results)

    def dump_pool(self):
        if self.pool:
            self.pool.close()
            self.pool = None

    def __null_reducer(self, vals):
        return vals

    def __map_serial(self, mapfunc, collection):
        for v in collection:
            yield mapfunc(v)
    
    def __map_pool(self, mapfunc, collection, pool):
        return pool.imap(mapfunc, collection)
    
    def __map_multiprocessed(self, mapfunc, collection):
        from multiprocessing import Pool, cpu_count
        procs = self.num_threads or (min(cpu_count(), 16) if cpu_count() else 1)
        if not self.pool:
            self.pool = Pool(procs)
        return self.__map_pool(mapfunc, collection, self.pool)
    
    def __map_threaded(self, mapfunc, collection):
        from multiprocessing.pool import ThreadPool
        from multiprocessing import cpu_count
        procs = self.num_threads or (min(cpu_count(), 16) if cpu_count() else 1)
        if not self.pool:
            self.pool = ThreadPool(procs)
        return self.__map_pool(mapfunc, collection, self.pool)



from setuptools import setup, Extension, find_packages

setup(name='yaul',
      version='0.4.3',
      author='Filip Sufitchi',
      author_email="fsufitchi@gmail.com",
      description="Yet Another Utility Library",
      url="https://github.com/fsufitch/YAUL",
      packages=['yaul'],
      package_dir={'':'src'},
      entry_points = {
        },

      install_requires=['contextlib2', 'python-daemon', 'psutil'],
      )
